'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function funnel (value, _ref) {
  var min = _ref[0],
      max = _ref[1];
  return value > max ? max : value < min ? min : value;
}

var interpolate = (function (_ref) {
  var _ref$completionRatio = _ref.completionRatio,
      completionRatio = _ref$completionRatio === void 0 ? 0.5 : _ref$completionRatio,
      _ref$range = _ref.range,
      to = _ref$range[0],
      from = _ref$range[1];
  var value = from + completionRatio * (to - from);
  /*
  console.log('||| to', to, 'from', from)
  console.log("||| completionRatio", completionRatio)
  console.log("||| value", value)
  */

  return value;
});

var miniHexToHex = function miniHexToHex(miniHex) {
  var miniHexArray = miniHex.split('');
  var hexTemplate = [0, 0, 1, 1, 2, 2];
  return hexTemplate.map(function (idx) {
    return miniHexArray[idx];
  }).join('');
};

var validateHex = {
  proto: '[a-fA-F0-9]+',
  draft: function draft(string) {
    var r = new RegExp("^" + this.proto + "$");
    return r.test(string) && string.length <= 6;
  },
  final: function final(string) {
    return string.length === 3 || string.length === 6;
  },
  exec: function exec(string) {
    var r = new RegExp(this.proto);
    return r.exec(string)[0];
  },
  process: function process(input) {
    var hex = this.exec(input);

    switch (hex.length) {
      case 6:
        return "#" + hex;

      case 3:
        return "#" + miniHexToHex(hex);

      default:
        throw new Error(hex + " does not contain a valid hex");
    }
  }
};

function wrapAround(value, _ref) {
  var min = _ref[0],
      max = _ref[1];

  while (value >= max) {
    value -= max - min;
  }

  while (value < min) {
    value += max - min;
  }

  return value;
}

/**
 * Gives us the relative values of the channels,
 * irrespective of the white light beneath them.
 * @param {number} hue - in degrees. Gets safely wrapped around first thing.
 *
 * DgiRed    0 ->   0
 *
 * Sunlgt   50 ->  50
 *
 * DgiYlw   60 ->  60
 *
 * Citron   70 ->  70
 *
 * Turqse  510 -> 150
 *
 * @const {number} hueReduced - hue 0-359.9 is now a floating point 0-5.999.
 *
 * DgiRed    0 ->   0  ~  0.000
 *
 * Sunlgt   50 -> 5/6  ~  0.833
 *
 * DgiYlw   60 ->   1  ~  1.000
 *
 * Citron   70 -> 7/6  ~  1.167
 *
 * Turqse  150 -> 5/2  ~  2.500
 *
 * @const {number} hueInteger - from 1-6. Tells us what color region we are in.
 *
 * DgiRed  0.000 -> 0
 *                : red-into-yellow region
 *
 * Sunlgt  0.833 -> 0
 *                : red-into-yellow region
 *
 * DgiYlw  1.000 -> 1
 *                : yellow-into-green region
 *
 * Citron  1.167 -> 1
 *                : yellow-into-green region
 *
 * Turqse  2.500 -> 2
 *                : green-into-cyan region
 *
 * hueInteger is the 'whole number' piece of hueReduced.
 * FYI, the six color regions are bounded by red, yellow, green, cyan, blue, magenta.
 *
 * @const {number} hueDecimal - tells where we are in this region.
 *
 * DgiRed  0.000 -> 0.000
 *                : at the very beginning
 *
 * Sunlgt  0.833 -> 0.833
 *                : near the end
 *
 * DgiYlw  1.000 -> 0.000
 *                : at the very beginning
 *
 * Citron  1.167 -> 0.167
 *                : near the beginning
 *
 * Turqse  2.500 -> 0.500
 *                : at the halfway point
 *
 * hueDecimal is the 'fraction' piece of hueReduced.
 * we are going to use this number to determine the value of the in-between channel.
 *
 * @const {number} x - used in primary-secondary transitions like Red into Yellow
 * @const {number} y - used in secondary-primary transitions like Yellow into Green
   * To understand the function of x and y, take the difference between
   *
   * Sunlgt (hue 50),
   * Citron (hue 70),
   * DgiYlw (hue 60),
   *
   * as an instructive case. These colors are all basically yellow.
   *
   * Sunlgt is hue 50, which puts it near the end of the red-into-yellow region.
     *
     * This means its Red channel is full, and its Green channel is almost full.
     * The fullness of its Green channel is directly proportional to its
     * hueDecimal, the distance from the beginning of this region: 0.833
     *
   * Citron is hue 70, which puts it near the beginning of the yellow-into-green region.
     *
     * This means its Red channel is ALMOST FULL, and its Green channel is FULL.
     * So the fullness of its Red channel is INVERSELY porportional to its
     * hueDecimal, the distance from the beginning of this region: 1 - 0.167 = 0.833
     *
   * DgiYlw is hue 60, which puts it at the very beginning of the yellow-into-green region.
     *
     * This means its Red Channel and its Green channel must both be full.
     * Like Citron, the fullness of DgiYlw's Red channel is inversely proporional to its
     * hueDecimal, which is 0. Therefore DgiYlw's Red channel has a fullness of 1.
   *
 * @returns array of values reflecting the spread between channels
 *
 * DgiRed  case 0:  [   R ===== 1       G = x = 0.000   B ===== 0      ]
 *
 * Sunlgt  case 0:  [   R ===== 1       G = x = 0.833   B ===== 0      ]
 *
 * DgiYlw  case 1:  [   R = y = 1.000   G ===== 1       B ===== 0      ]
 *
 * Citron  case 1:  [   R = y = 0.833   G ===== 1       B ===== 0      ]
 *
 * Turqse  case 2:  [   R ===== 0       G ===== 1       B = x = 0.500  ]
 *
 * here we see detailed breakdowns of the function's final output for our running examples.
 */

var hueToRelativeChannels = (function (hue) {
  hue = wrapAround(hue, [0, 360]);
  var hueReduced = hue / 60;
  var hueInteger = Math.floor(hueReduced);
  var hueDecimal = hueReduced - hueInteger;
  var x = hueDecimal;
  var y = 1 - hueDecimal;

  switch (hueInteger) {
    case 0:
      return [1, x, 0];

    case 1:
      return [y, 1, 0];

    case 2:
      return [0, 1, x];

    case 3:
      return [0, y, 1];

    case 4:
      return [x, 0, 1];

    case 5:
      return [1, 0, y];

    default:
      throw new Error("invalid hue served: " + hue);
  }
});

var HUE_STRUCTURES = {
  none: [],
  polar: [180],
  split: [150, 210],
  trine: [120, 240],
  square: [90, 180, 270]
};
var CHANNEL_SPECIFIC_LUM = {
  R: 0.3,
  G: 0.5,
  B: 0.2
};
/*
const I'M_WONDERING_IF_THIS_IS_NOT_VERY_ACCURATE = {
  R: 0.2126,
  G: 0.7152,
  B: 0.0722,
}
*/

var specificLumFromHue = (function (hue) {
  var _hueToRelativeChannel = hueToRelativeChannels(hue),
      factorR = _hueToRelativeChannel[0],
      factorG = _hueToRelativeChannel[1],
      factorB = _hueToRelativeChannel[2];

  var lumR = CHANNEL_SPECIFIC_LUM.R * factorR;
  var lumG = CHANNEL_SPECIFIC_LUM.G * factorG;
  var lumB = CHANNEL_SPECIFIC_LUM.B * factorB;
  var specificLum = lumR + lumG + lumB;
  return specificLum;
});

var maxSatForHueFromTuner = (function (hue, tuner) {
  // 430
  var maxSat = 255;
  var hueWrapped = wrapAround(hue, [0, 360]); // 70

  for (var a = -1, b = 0; b < tuner.length; a++, b++) {
    a = wrapAround(a, [0, tuner.length]); // console.log('||| a =', a, 'b =', b)

    var hueDoubleWrapped = a > b ? wrapAround(hueWrapped, [-180, 180]) : undefined; // undef

    var tuningPointA = tuner[a];
    var tuningPointB = tuner[b];
    var hueA = a > b ? wrapAround(tuningPointA.hue, [-180, 180]) : tuningPointA.hue;
    var hueB = tuningPointB.hue;

    if ((hueDoubleWrapped || hueWrapped) >= hueA && (hueDoubleWrapped || hueWrapped) < hueB) {
      // console.log('||| hue', hue, 'is between', tuningPointA.hue, 'and', tuningPointB.hue)
      var $ = hueDoubleWrapped || hueWrapped; // 70

      $ -= hueA; // 70 - 50 = 20 //

      $ /= hueB - hueA; // 20 / (120 - 50) = 2/7

      $ *= tuningPointB.sat - tuningPointA.sat; // -128 * 2 / 7 = -256 / 7 ~= -37

      $ += tuningPointA.sat;

      maxSat = $;
    }
  } // console.log('--- maxSat', maxSat)


  return maxSat;
});

var hueFromChannels = (function (_ref) {
  var R = _ref.R,
      G = _ref.G,
      B = _ref.B;
  var hue;
  if (R > G && G >= B) hue = 60 * (0 + (G - B) / (R - B));
  if (G >= R && R > B) hue = 60 * (2 - (R - B) / (G - B));
  if (G > B && B >= R) hue = 60 * (2 + (B - R) / (G - R));
  if (B >= G && G > R) hue = 60 * (4 - (G - R) / (B - R));
  if (B > R && R >= G) hue = 60 * (4 + (R - G) / (B - G));
  if (R >= B && B > G) hue = 60 * (6 - (B - G) / (R - G));
  if (R === G && G === B) hue = 0; // console.log('||| found hue', hue)

  return hue;
});

var satFromChannels = (function (_ref) {
  var R = _ref.R,
      G = _ref.G,
      B = _ref.B;
  var sat = Math.max(R, G, B) - Math.min(R, G, B); // console.log('||| found sat', sat)

  return sat;
});

var lumFromChannels = (function (_ref) {
  var R = _ref.R,
      G = _ref.G,
      B = _ref.B;
  var lum = CHANNEL_SPECIFIC_LUM.R * R / 255 + CHANNEL_SPECIFIC_LUM.G * G / 255 + CHANNEL_SPECIFIC_LUM.B * B / 255; // console.log('||| found lum', lum)

  return lum;
});

var specToChannels = (function (_ref) {
  var hue = _ref.hue,
      sat = _ref.sat,
      lum = _ref.lum,
      _ref$prefer = _ref.prefer,
      prefer = _ref$prefer === void 0 ? 'sat' : _ref$prefer,
      tuner = _ref.tuner;

  /*
  console.log('||| hue', hue)
  console.log('||| sat', sat)
  console.log('||| lum', lum)
  console.log('||| prefer', prefer)
  */
  var relativeChannels = hueToRelativeChannels(hue);

  function makeMinChannels(trueSaturation) {
    var makeMinChannel = function makeMinChannel(idx) {
      return Math.round(relativeChannels[idx] * trueSaturation);
    };

    return {
      R: makeMinChannel(0),
      G: makeMinChannel(1),
      B: makeMinChannel(2)
    };
  }

  var trueSaturation;
  var trueLuminosity;
  var minChannels;
  var minLum = 0;
  var maxLum = 1;
  var maxSat = maxSatForHueFromTuner(hue, tuner);

  if (prefer === 'sat') {
    trueSaturation = funnel(Math.min(sat, maxSat), [0, 255]);
    minChannels = makeMinChannels(trueSaturation);
    var maxChannels = {
      R: minChannels.R + 255 - trueSaturation,
      G: minChannels.G + 255 - trueSaturation,
      B: minChannels.B + 255 - trueSaturation
    };
    minLum = lumFromChannels(minChannels);
    maxLum = lumFromChannels(maxChannels);
    trueLuminosity = funnel(lum, [minLum, maxLum]);
    /*
    console.log('||| trueSaturation', trueSaturation)
    console.log('||| idealLum', lum)
    console.log('||| minChannels', minChannels)
    console.log('||| minLum', minLum)
    console.log('||| maxChannels', maxChannels)
    console.log('||| maxLum', maxLum)
    console.log('||| trueLuminosity', trueLuminosity)
    */
  }

  if (prefer === 'lum') {
    trueLuminosity = funnel(lum, [0, 1]);
    var specificLum = specificLumFromHue(hue);
    maxSat = Math.min(maxSat, Math.round(trueLuminosity <= specificLum ? 255 * (trueLuminosity / specificLum) : 255 * (1 - trueLuminosity) / (1 - specificLum)));
    trueSaturation = Math.min(sat, maxSat);
    minChannels = makeMinChannels(trueSaturation);
    minLum = lumFromChannels(minChannels);
    /*
    console.log('||| trueLuminosity', trueLuminosity)
    console.log('||| lumAtSat255', lumAtSat255)
    console.log('||| maxSat', maxSat)
    console.log('||| trueSaturation', trueSaturation)
    console.log('||| minChannels', minChannels)
    */
  }

  var maxWhite = 255 - Math.max.apply(Math, Object.values(minChannels));
  var white = funnel(Math.round((trueLuminosity - minLum) * 255), [0, maxWhite]);
  var _R$G$B = {
    R: minChannels.R + white,
    G: minChannels.G + white,
    B: minChannels.B + white
  },
      R = _R$G$B.R,
      G = _R$G$B.G,
      B = _R$G$B.B;
  /*
  console.log('||| maxWhite', maxWhite)
  console.log('||| white', white)
  console.log('||| R', R)
  console.log('||| G', G)
  console.log('||| B', B)
  */

  return {
    channels: {
      R: R,
      G: G,
      B: B
    },
    fix: {
      sat: trueSaturation,
      lum: trueLuminosity
    },
    limit: {
      sat: [0, maxSat],
      lum: [prefer === 'lum' ? 0 : minLum, maxLum]
    }
  };
});

var channelsToHex = (function (channels) {
  return "#" + Object.values(channels).map(function (channel) {
    var channelHex = channel.toString(16);
    if (channelHex.length === 1) channelHex = 0 + channelHex;
    return channelHex;
  }).join('');
});

var specToHexFixLimit = (function (_ref) {
  var hue = _ref.hue,
      sat = _ref.sat,
      lum = _ref.lum,
      prefer = _ref.prefer,
      tuner = _ref.tuner;

  var _specToChannels = specToChannels({
    hue: hue,
    sat: sat,
    lum: lum,
    prefer: prefer,
    tuner: tuner
  }),
      channels = _specToChannels.channels,
      fix = _specToChannels.fix,
      limit = _specToChannels.limit;

  var R = channels.R,
      G = channels.G,
      B = channels.B;
  var hex = channelsToHex({
    R: R,
    G: G,
    B: B
  }); // console.log('--- newHex', hex)

  return {
    hex: hex,
    fix: fix,
    limit: limit
  };
});

var specToHex = (function (_ref) {
  var hue = _ref.hue,
      sat = _ref.sat,
      lum = _ref.lum,
      prefer = _ref.prefer,
      tuner = _ref.tuner;

  var _specToHexFixLimit = specToHexFixLimit({
    hue: hue,
    sat: sat,
    lum: lum,
    prefer: prefer,
    tuner: tuner
  }),
      hex = _specToHexFixLimit.hex;

  return hex;
});

function interpolateChanges(rootSpec, currentStep, axes) {
  var changes = {};

  var _loop = function _loop(axisIdx) {
    var currentAxis = axes[axisIdx];
    var attribute = currentAxis.attribute,
        relative = currentAxis.relative;
    var rootValue = rootSpec[attribute];

    var to = function to() {
      switch (typeof currentAxis.to) {
        case 'number':
          if (!relative) return currentAxis.to;
          return rootValue + currentAxis.to;

        case 'function':
          return currentAxis.to(rootSpec[attribute]);

        default:
          throw new Error("gradient axis type error:\n" + ("currentAxis.to is " + currentAxis.to + " (type: " + typeof currentAxis.to + ")\n") + "type must be\n" + "* 'number'    (for absolute values) or\n" + "* 'function'  (for values relative to the root value).");
      }
    };

    var from = function from() {
      switch (typeof currentAxis.from) {
        case 'undefined':
          return rootSpec[attribute];

        case 'number':
          if (!relative) return currentAxis.from;
          return rootValue + currentAxis.from;

        case 'function':
          return currentAxis.from(rootSpec[attribute]);

        default:
          throw new Error("gradient axis type error:\n" + ("currentAxis.from is " + currentAxis.from + " (type: " + typeof currentAxis.from + ")\n") + "type must be\n" + "* 'undefined' (to default to the root value)\n" + "* 'number'    (for absolute values) or\n" + "* 'function'  (for values relative to the root value).");
      }
    };

    var value = interpolate({
      completionRatio: currentStep,
      range: [to(), from()]
    });
    changes[attribute] = value;
  };

  for (var axisIdx = 0; axisIdx < axes.length; axisIdx++) {
    _loop(axisIdx);
  } // console.log('>>> changes', changes)


  return changes;
}

function gradientToHexArray(rootSpec, gradient, tuner) {
  var steps = gradient.steps,
      axes = gradient.axes,
      prefer = gradient.prefer;
  var hexArray = [];

  for (var stepIdx = 0; stepIdx < steps.length; stepIdx++) {
    var currentStep = steps[stepIdx];
    var changes = interpolateChanges(rootSpec, currentStep, axes);
    var hex = specToHex(_extends({}, rootSpec, changes, {
      prefer: prefer,
      tuner: tuner
    }));
    hexArray.push(hex);
  }

  return hexArray;
}

var gradientsToHexArrays = (function (originalColorObject, tuner) {
  var gradients = originalColorObject.gradients;
  var hue = originalColorObject.hue,
      sat = originalColorObject.sat,
      lum = originalColorObject.lum;
  var rootSpec = {
    hue: hue,
    sat: sat,
    lum: lum
  };
  var hexArrays = [];

  for (var gradientIdx = 0; gradientIdx < gradients.length; gradientIdx++) {
    var gradient = gradients[gradientIdx];
    var hexArray = gradientToHexArray(rootSpec, gradient, tuner);
    hexArrays.push(hexArray);
  }

  return hexArrays;
});

var channelsToSpec = (function (_ref) {
  var R = _ref.R,
      G = _ref.G,
      B = _ref.B;
  var hue = hueFromChannels({
    R: R,
    G: G,
    B: B
  });
  var sat = satFromChannels({
    R: R,
    G: G,
    B: B
  });
  var lum = lumFromChannels({
    R: R,
    G: G,
    B: B
  });
  return {
    hue: hue,
    sat: sat,
    lum: lum
  };
});

var hexToChannels = (function (hex) {
  var validated = validateHex.process(hex);

  var channel = function channel(idx) {
    return parseInt(validated.substr(idx * 2 + 1, 2), 16);
  };

  return {
    R: channel(0),
    G: channel(1),
    B: channel(2)
  };
});

var hexToSpec = (function (hex) {
  var _hexToChannels = hexToChannels(hex),
      R = _hexToChannels.R,
      G = _hexToChannels.G,
      B = _hexToChannels.B;
  /*
  console.log('+++ input hex', hex)
  console.log('||| R', R)
  console.log('||| G', G)
  console.log('||| B', B)
  */


  var _channelsToSpec = channelsToSpec({
    R: R,
    G: G,
    B: B
  }),
      hue = _channelsToSpec.hue,
      sat = _channelsToSpec.sat,
      lum = _channelsToSpec.lum;

  return {
    hue: hue,
    sat: sat,
    lum: lum
  };
});

function mixPalettes(_ref) {
  var palettes = _ref.palettes,
      hexes = _ref.hexes,
      ops = _ref.ops,
      _ref$place = _ref.place;
  _ref$place = _ref$place === void 0 ? [0, null, null] : _ref$place;
  var paletteIdx = _ref$place[0],
      stateKey = _ref$place[1],
      elementKey = _ref$place[2];
  if (!Array.isArray(palettes) || palettes.length === 0) throw new Error('Must provide at least one palette');
  var palettes_ = [].concat(palettes);
  var palette = palettes_[paletteIdx];
  var state = stateKey ? palette.states[stateKey] : null;
  var element = elementKey ? state.elements[elementKey] : null;
  var destination = element || state || palette;
  var origin = element ? state : palette;
  if (!origin && hexes.length === 0) throw new Error('Must provide at least one hex');
  if (!origin.color) origin.color = hexToSpec(hexes[0]);

  var color = _extends({}, origin.color);

  var opsEntries = Object.entries(ops);
  /*
  console.log("--- from", { ...origin }.id, 'to', { ...destination }.id,
    `[${paletteIdx}, ${stateKey}, ${elementKey}]`,
  )
  console.log("    ||| origin", origin)
  console.log("    ||| hexes", hexes)
  console.log("    ||| color", color)
  console.log("    ||| ops", ops)
  console.log("    ||| palette", [...palette])
  */

  for (var idx = 0; idx < opsEntries.length; idx++) {
    var _opsEntries$idx = opsEntries[idx],
        key = _opsEntries$idx[0],
        value = _opsEntries$idx[1];
    /*
    console.log("    > color", ...color)
    console.log('     ', key)
    console.log('     ', value)
    */

    switch (key) {
      case 'color':
        // change colors
        // capture currentColor in palette
        switch (typeof value) {
          case 'number':
            color = hexToSpec(hexes[value]);
            break;

          case 'string':
            hexes.unshift(value);
            color = hexToSpec(value);
            break;

          default:
            throw new Error("'color' accepts types 'string' or 'number'; got type '" + typeof value + "'");
        }

        break;

      case 'hue':
      case 'sat':
      case 'lum':
        switch (typeof value) {
          case 'number':
            color[key] = value;
            break;

          case 'function':
            color[key] = value(color[key]);
            break;

          default:
            throw new Error("hue, sat, lum accept types 'number' or 'function'; got type '" + typeof value + "'");
        }

        break;

      case 'contrast':
        switch (value) {
          case 'hard':
            color.lum = color.lum > 0.67 ? 0 : 1;
            break;

          case 'soft':
            color.lum = color.lum > 0.67 ? 0.05 : 0.95;
            break;

          case 'harden':
            color.lum = color.lum > 0.67 ? 0 : 1;
            break;

          default:
            throw new Error("'contrast' accepts 'hard' or 'soft'; got '" + value + "'");
        }

        color.prefer = 'lum';
        break;

      case 'split':
      case 'trine':
      case 'tetra':
        break;

      case 'warm':
      case 'cool':
        break;

      case 'amp':
      case 'mute':
        break;

      case 'tint':
      case 'shade':
        var percentSaturated = color.sat / 255;
        var specificLum = specificLumFromHue(color.hue);
        var lightAdjust = 0.5 * (1 - percentSaturated) + specificLum * percentSaturated;
        var charge = key === 'tint' ? 100 : -100;
        var offset = lightAdjust * value / charge; // console.log('lightAdjust', lightAdjust)
        // console.log('offset', offset)

        color.lum += offset;
        if (color.lum < 0 || color.lum > 1) color.lum -= 2.5 * offset;
        color.prefer = 'lum';
        break;

      case 'holds':
        var holdsEntries = Object.entries(value);

        for (var holdsEntryIdx = 0; holdsEntryIdx < holdsEntries.length; holdsEntryIdx++) {
          var _holdsEntries$holdsEn = holdsEntries[holdsEntryIdx],
              subPaletteId = _holdsEntries$holdsEn[0],
              subPaletteOps = _holdsEntries$holdsEn[1];
          var externalElements = {};
          var baseElementsEntries = Object.entries(palette.states.base.elements);

          for (var baseElementsIdx = 0; baseElementsIdx < baseElementsEntries.length; baseElementsIdx++) {
            var _baseElementsEntries$ = baseElementsEntries[baseElementsIdx],
                baseElementKey = _baseElementsEntries$[0],
                baseElementValue = _baseElementsEntries$[1];
            externalElements["ex-" + baseElementKey] = baseElementValue;
          }

          palettes_.push({
            id: subPaletteId,
            color: palettes_[0].color,
            states: {
              base: {
                id: 'base',
                elements: externalElements
              }
            }
          });
          palettes_ = mixPalettes({
            palettes: palettes_,
            hexes: hexes,
            ops: subPaletteOps,
            place: [palettes_.length - 1, null, null]
          });
        }

        break;

      case 'state':
        var stateEntries = Object.entries(value);

        if (!element) {
          for (var stateEntryIdx = 0; stateEntryIdx < stateEntries.length; stateEntryIdx++) {
            var _stateEntries$stateEn = stateEntries[stateEntryIdx],
                stateKey_ = _stateEntries$stateEn[0],
                stateOps = _stateEntries$stateEn[1]; // console.log('    ||| stateEntry', stateKey_, stateOps)

            palette.states[stateKey_] = {
              id: stateKey_,
              elements: {}
            };
            palette.states[stateKey_].color = mixPalettes({
              palettes: palettes,
              hexes: hexes,
              ops: stateOps,
              place: [paletteIdx, stateKey_, null]
            }); // console.log('      ', palette.states[stateKey_].hex)
          }
        } else {
          for (var _stateEntryIdx = 0; _stateEntryIdx < stateEntries.length; _stateEntryIdx++) {
            var _stateEntries$_stateE = stateEntries[_stateEntryIdx],
                _stateKey_ = _stateEntries$_stateE[0],
                _stateOps = _stateEntries$_stateE[1]; // console.log('    ||| >>> stateEntry', stateKey_, stateOps)
            // console.log('    ||| >>> elementKey', elementKey)

            var state_ = palette.states[_stateKey_];
            if (!state_.elements[elementKey]) state_.elements[elementKey] = {};
            palette.states[_stateKey_].elements[elementKey].hex = mixPalettes({
              palettes: palettes,
              hexes: hexes,
              ops: _stateOps,
              place: [paletteIdx, _stateKey_, elementKey]
            });
          }
        }

        break;

      default:
        var elementKey_ = key;
        var stateKeys = Object.keys(palette.states);

        for (var stateKeysIdx = 0; stateKeysIdx < stateKeys.length; stateKeysIdx++) {
          var _stateKey_2 = stateKeys[stateKeysIdx];
          var _state_ = palette.states[_stateKey_2]; // console.log('    ||| state_', state_)

          _state_.elements[elementKey_] = {
            id: elementKey_
          };
          _state_.elements[elementKey_].color = mixPalettes({
            palettes: palettes,
            hexes: hexes,
            ops: value,
            place: [paletteIdx, _stateKey_2, elementKey_]
          });
        }

    } // hex = specToHex({ ...color, tuner })


    destination.color = color;
    if (!state) destination.states.base.color = color; // console.log("    < color", color, color)
    // console.log("      set", { ...destination }.id)
  }

  if (element || state) {
    // console.log('---', destination.id, 'color', color)
    return color;
  } // console.log('--- returning palettes_', palettes_)


  return palettes_;
}

function paletteToScssBlock(palette, tuner) {
  var scssBlock = '';
  var baseEntries = Object.entries(palette.states.base.elements);

  for (var baseIdx = 0; baseIdx < baseEntries.length; baseIdx++) {
    var _baseEntries$baseIdx = baseEntries[baseIdx],
        key = _baseEntries$baseIdx[0],
        element = _baseEntries$baseIdx[1];
    scssBlock += "  --" + key + "-color: " + specToHex(_extends({}, element.color, {
      tuner: tuner
    })) + ";\n";
  }

  var stateEntries = Object.entries(palette.states);

  for (var stateIdx = 0; stateIdx < stateEntries.length; stateIdx++) {
    var _stateEntries$stateId = stateEntries[stateIdx],
        stateKey = _stateEntries$stateId[0],
        stateObj = _stateEntries$stateId[1]; // console.log('stateKey', stateKey)

    switch (stateKey) {
      case 'hover':
        scssBlock += "  &:hover, &:focus, &:focus-within {\n";
        break;

      case 'active':
        scssBlock += "  &:active, &.active {\n";
        break;

      case 'disabled':
        scssBlock += "  &:disabled, &.disabled {\n";
        break;

      default:
        continue;
    }

    var elementEntries = Object.entries(stateObj.elements);

    for (var elementIdx = 0; elementIdx < elementEntries.length; elementIdx++) {
      var _elementEntries$eleme = elementEntries[elementIdx],
          _key = _elementEntries$eleme[0],
          _element = _elementEntries$eleme[1];
      scssBlock += "    --" + _key + "-color: " + specToHex(_extends({}, _element.color, {
        tuner: tuner
      })) + ";\n";
    }

    scssBlock += '  }\n';
  }

  return scssBlock;
}

function makeScssObject(palettes, tuner) {
  var scssObject = {};

  for (var index = 0; index < palettes.length; index++) {
    var palette = palettes[index];
    scssObject[palette.id] = paletteToScssBlock(palette, tuner);
  }

  return scssObject;
}

var paletteTemplate = function paletteTemplate(name) {
  return {
    id: name,
    states: {
      base: {
        elements: {}
      }
    }
  };
};

function makeColorPalettesFromHexes(_ref2) {
  var hexes = _ref2.hexes,
      scheme = _ref2.scheme,
      tuner = _ref2.tuner;
  var schemeEntries = Object.entries(scheme);
  var palettes = [];

  for (var idx = 0; idx < schemeEntries.length; idx++) {
    var _schemeEntries$idx = schemeEntries[idx],
        name = _schemeEntries$idx[0],
        ops = _schemeEntries$idx[1];
    var newPalettes = mixPalettes({
      palettes: [paletteTemplate(name)],
      hexes: hexes,
      ops: ops
    });
    palettes = [].concat(palettes, newPalettes);
  }

  var scss = makeScssObject(palettes, tuner); // console.log(scss)

  return scss;
}

var getStandout = (function (hex) {
  var _hexToSpec = hexToSpec(hex),
      lum = _hexToSpec.lum;

  var standOut = lum > 0.666 ? 'black' : 'white';
  return standOut;
}); // input: red
// hard standout: white

var getOffset = (function (_ref) {
  var hex = _ref.hex,
      tuner = _ref.tuner,
      offsets = _ref.offsets;
  var rootSpec = hexToSpec(hex);
  var changes = {};

  for (var index = 0; index < offsets.length; index++) {
    var _offsets$index = offsets[index],
        attribute = _offsets$index.attribute,
        offsetValue = _offsets$index.offsetValue;
    var rootValue = rootSpec[attribute];
    var sat = rootSpec.sat;
    var newValue = void 0;

    switch (attribute) {
      case 'hue':
      case 'sat':
        newValue = rootValue + offsetValue;
        break;

      case 'lum':
        if (offsetValue + rootValue < 0 || offsetValue + rootValue > 1) {
          newValue = rootValue - (offsetValue - (sat >= 170 ? offsetValue * (sat - 170) / 85 : 0)) * 1.67;
        } else {
          newValue = rootValue + (offsetValue - (sat >= 170 ? 0.333 * offsetValue * (sat - 170) / 85 : 0));
        }

        break;

      default:
        throw new Error('Received spec attribute other than "hue," "sat," or "lum."');
    }

    changes[attribute] = newValue;
  }

  var newHex = specToHex(_extends({}, rootSpec, changes, {
    prefer: 'lum',
    tuner: tuner
  }));
  return newHex;
});

function hardContrast(hex) {
  var _hexToSpec = hexToSpec(hex),
      lum = _hexToSpec.lum;

  var contrast = lum > 0.666 ? 'black' : 'white';
  return contrast;
}
function softContrast(hex, tuner) {
  var rootSpec = hexToSpec(hex);
  var newLum = rootSpec.lum > 0.666 ? 0.05 : 0.95;
  var newHex = specToHex(_extends({}, rootSpec, {
    lum: newLum,
    prefer: 'lum',
    tuner: tuner
  }));
  return newHex;
} // input: red
// hard standout: white

var rangeNames = [{
  hue: 20,
  name: 'orange'
}, {
  hue: 45,
  name: 'yellow'
}, {
  hue: 61,
  name: 'citron'
}, {
  hue: 80,
  name: 'lime'
}, {
  hue: 100,
  name: 'green'
}, {
  hue: 150,
  name: 'teal'
}, {
  hue: 180,
  name: 'cyan'
}, {
  hue: 200,
  name: 'blue'
}, {
  hue: 240,
  name: 'indigo'
}, {
  hue: 270,
  name: 'violet'
}, {
  hue: 290,
  name: 'magenta'
}, {
  hue: 330,
  name: 'pink'
}, {
  hue: 350,
  name: 'red'
}];
var nameHue = (function (hue) {
  console.log('||| hue', hue);
  var hueWrapped = wrapAround(hue, [0, 360]);

  for (var a = -1, b = 0; b < rangeNames.length; a++, b++) {
    a = wrapAround(a, [0, rangeNames.length]);
    var hueDoubleWrapped = a > b ? wrapAround(hueWrapped, [-180, 180]) : undefined;
    var namingPointA = rangeNames[a];
    var namingPointB = rangeNames[b];
    var hueA = a > b ? wrapAround(namingPointA.hue, [-180, 180]) : namingPointA.hue;
    var hueB = namingPointB.hue;

    if ((hueDoubleWrapped || hueWrapped) >= hueA && (hueDoubleWrapped || hueWrapped) < hueB) {
      console.log('||| hue', hue, 'is between', namingPointA.hue, 'and', namingPointB.hue, 'therefore it is named', namingPointA.name);
      return namingPointA.name;
    }
  }
});

var builtInTunerKit = {
  noFilter: [{
    sat: 255,
    hue: 0
  }, {
    sat: 255,
    hue: 360
  }],
  simulateCMYK: [{
    sat: 255,
    hue: 55
  }, {
    sat: 127,
    hue: 120
  }, {
    sat: 255,
    hue: 200
  }, {
    sat: 255,
    hue: 205
  }, {
    sat: 170,
    hue: 240
  }, {
    sat: 230,
    hue: 270
  }, {
    sat: 210,
    hue: 300
  }, {
    sat: 255,
    hue: 320
  }]
};

var schemes = {
  trifactory: {
    PaletteModule: {
      color: 0,
      bg: {},
      mg: {
        shade: 10
      },
      fg: {
        contrast: 'soft'
      },
      holds: {
        Control: {
          state: {
            hover: {},
            active: {},
            disabled: {}
          },
          bg: {
            shade: 24,
            state: {
              base: {},
              hover: {
                shade: 18
              },
              active: {
                shade: 32
              },
              disabled: {
                shade: 32
              }
            }
          },
          mg: {
            shade: 48
          },
          fg: {
            contrast: 'soft',
            state: {
              hover: {
                contrast: 'harden'
              },
              active: {
                contrast: 'harden'
              },
              disabled: {
                shade: 20
              }
            }
          }
        }
      }
    }
  }
}; // Example:
/*
const colorSchema = {
  exfg: colorContext.softContrast,
  exbg: hex,
  fg: {
    base: colorContext.softContrast,
    hover: colorContext.softContrast,
    active: colorContext.hardContrast,
    disable: a,
  },
  mg: [20, 15, 25].map(n => shade(n)),
  bg: [10, 5, 15].map(n => shade(n)),
}

const stylesheet = `
a
`
*/

exports.CHANNEL_SPECIFIC_LUM = CHANNEL_SPECIFIC_LUM;
exports.HUE_STRUCTURES = HUE_STRUCTURES;
exports.builtInTunerKit = builtInTunerKit;
exports.channelsToSpec = channelsToSpec;
exports.funnel = funnel;
exports.getOffset = getOffset;
exports.getStandout = getStandout;
exports.gradientsToHexArrays = gradientsToHexArrays;
exports.hardContrast = hardContrast;
exports.hexToChannels = hexToChannels;
exports.hexToSpec = hexToSpec;
exports.hueFromChannels = hueFromChannels;
exports.hueToRelativeChannels = hueToRelativeChannels;
exports.interpolate = interpolate;
exports.lumFromChannels = lumFromChannels;
exports.maxSatForHueFromTuner = maxSatForHueFromTuner;
exports.nameHue = nameHue;
exports.satFromChannels = satFromChannels;
exports.scheme = makeColorPalettesFromHexes;
exports.schemes = schemes;
exports.softContrast = softContrast;
exports.specToHex = specToHex;
exports.specToHexFixLimit = specToHexFixLimit;
exports.specificLumFromHue = specificLumFromHue;
exports.validateHex = validateHex;
exports.wrapAround = wrapAround;
//# sourceMappingURL=luum.js.map
