'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function funnel (value, [min, max]) {
  return value > max ? max : value < min ? min : value;
}

var interpolate = (({
  completionRatio = 0.5,
  range: [to, from]
}) => {
  const value = from + completionRatio * (to - from);
  /*
  console.log('||| to', to, 'from', from)
  console.log("||| completionRatio", completionRatio)
  console.log("||| value", value)
  */

  return value;
});

const miniHexToHex = miniHex => {
  const miniHexArray = miniHex.split('');
  const hexTemplate = [0, 0, 1, 1, 2, 2];
  return hexTemplate.map(idx => miniHexArray[idx]).join('');
};

var validateHex = {
  proto: '[a-fA-F0-9]+',

  draft(string) {
    const r = new RegExp(`^${this.proto}$`);
    return r.test(string) && string.length <= 6;
  },

  final(string) {
    return string.length === 3 || string.length === 6;
  },

  exec(string) {
    const r = new RegExp(this.proto);
    return r.exec(string)[0];
  },

  process(input) {
    const hex = this.exec(input);

    switch (hex.length) {
      case 6:
        return `#${hex}`;

      case 3:
        return `#${miniHexToHex(hex)}`;

      default:
        throw new Error(`${hex} does not contain a valid hex`);
    }
  }

};

function wrapAround(value, [min, max]) {
  while (value >= max) value -= max - min;

  while (value < min) value += max - min;

  return value;
}

/**
 * Gives us the relative values of the channels,
 * irrespective of the white light beneath them.
 * @param {number} hue - in degrees. Gets safely wrapped around first thing.
 *
 * DgiRed    0 ->   0
 *
 * Sunlgt   50 ->  50
 *
 * DgiYlw   60 ->  60
 *
 * Citron   70 ->  70
 *
 * Turqse  510 -> 150
 *
 * @const {number} hueReduced - hue 0-359.9 is now a floating point 0-5.999.
 *
 * DgiRed    0 ->   0  ~  0.000
 *
 * Sunlgt   50 -> 5/6  ~  0.833
 *
 * DgiYlw   60 ->   1  ~  1.000
 *
 * Citron   70 -> 7/6  ~  1.167
 *
 * Turqse  150 -> 5/2  ~  2.500
 *
 * @const {number} hueInteger - from 1-6. Tells us what color region we are in.
 *
 * DgiRed  0.000 -> 0
 *                : red-into-yellow region
 *
 * Sunlgt  0.833 -> 0
 *                : red-into-yellow region
 *
 * DgiYlw  1.000 -> 1
 *                : yellow-into-green region
 *
 * Citron  1.167 -> 1
 *                : yellow-into-green region
 *
 * Turqse  2.500 -> 2
 *                : green-into-cyan region
 *
 * hueInteger is the 'whole number' piece of hueReduced.
 * FYI, the six color regions are bounded by red, yellow, green, cyan, blue, magenta.
 *
 * @const {number} hueDecimal - tells where we are in this region.
 *
 * DgiRed  0.000 -> 0.000
 *                : at the very beginning
 *
 * Sunlgt  0.833 -> 0.833
 *                : near the end
 *
 * DgiYlw  1.000 -> 0.000
 *                : at the very beginning
 *
 * Citron  1.167 -> 0.167
 *                : near the beginning
 *
 * Turqse  2.500 -> 0.500
 *                : at the halfway point
 *
 * hueDecimal is the 'fraction' piece of hueReduced.
 * we are going to use this number to determine the value of the in-between channel.
 *
 * @const {number} x - used in primary-secondary transitions like Red into Yellow
 * @const {number} y - used in secondary-primary transitions like Yellow into Green
   * To understand the function of x and y, take the difference between
   *
   * Sunlgt (hue 50),
   * Citron (hue 70),
   * DgiYlw (hue 60),
   *
   * as an instructive case. These colors are all basically yellow.
   *
   * Sunlgt is hue 50, which puts it near the end of the red-into-yellow region.
     *
     * This means its Red channel is full, and its Green channel is almost full.
     * The fullness of its Green channel is directly proportional to its
     * hueDecimal, the distance from the beginning of this region: 0.833
     *
   * Citron is hue 70, which puts it near the beginning of the yellow-into-green region.
     *
     * This means its Red channel is ALMOST FULL, and its Green channel is FULL.
     * So the fullness of its Red channel is INVERSELY porportional to its
     * hueDecimal, the distance from the beginning of this region: 1 - 0.167 = 0.833
     *
   * DgiYlw is hue 60, which puts it at the very beginning of the yellow-into-green region.
     *
     * This means its Red Channel and its Green channel must both be full.
     * Like Citron, the fullness of DgiYlw's Red channel is inversely proporional to its
     * hueDecimal, which is 0. Therefore DgiYlw's Red channel has a fullness of 1.
   *
 * @returns array of values reflecting the spread between channels
 *
 * DgiRed  case 0:  [   R ===== 1       G = x = 0.000   B ===== 0      ]
 *
 * Sunlgt  case 0:  [   R ===== 1       G = x = 0.833   B ===== 0      ]
 *
 * DgiYlw  case 1:  [   R = y = 1.000   G ===== 1       B ===== 0      ]
 *
 * Citron  case 1:  [   R = y = 0.833   G ===== 1       B ===== 0      ]
 *
 * Turqse  case 2:  [   R ===== 0       G ===== 1       B = x = 0.500  ]
 *
 * here we see detailed breakdowns of the function's final output for our running examples.
 */

var hueToRelativeChannels = (hue => {
  hue = wrapAround(hue, [0, 360]);
  const hueReduced = hue / 60;
  const hueInteger = Math.floor(hueReduced);
  const hueDecimal = hueReduced - hueInteger;
  const x = hueDecimal;
  const y = 1 - hueDecimal;

  switch (hueInteger) {
    case 0:
      return [1, x, 0];

    case 1:
      return [y, 1, 0];

    case 2:
      return [0, 1, x];

    case 3:
      return [0, y, 1];

    case 4:
      return [x, 0, 1];

    case 5:
      return [1, 0, y];

    default:
      throw new Error(`invalid hue served: ${hue}`);
  }
});

const HUE_STRUCTURES = {
  none: [],
  polar: [180],
  split: [150, 210],
  trine: [120, 240],
  square: [90, 180, 270]
};
const CHANNEL_SPECIFIC_LUM = {
  R: 0.3,
  G: 0.5,
  B: 0.2
};
/*
const I'M_WONDERING_IF_THIS_IS_NOT_VERY_ACCURATE = {
  R: 0.2126,
  G: 0.7152,
  B: 0.0722,
}
*/

var specificLumFromHue = (hue => {
  const [factorR, factorG, factorB] = hueToRelativeChannels(hue);
  const lumR = CHANNEL_SPECIFIC_LUM.R * factorR;
  const lumG = CHANNEL_SPECIFIC_LUM.G * factorG;
  const lumB = CHANNEL_SPECIFIC_LUM.B * factorB;
  const specificLum = lumR + lumG + lumB;
  return specificLum;
});

var maxSatForHueFromTuner = ((hue, tuner) => {
  // 430
  let maxSat = 255;
  const hueWrapped = wrapAround(hue, [0, 360]); // 70

  for (let a = -1, b = 0; b < tuner.length; a++, b++) {
    a = wrapAround(a, [0, tuner.length]); // console.log('||| a =', a, 'b =', b)

    const hueDoubleWrapped = a > b ? wrapAround(hueWrapped, [-180, 180]) : undefined; // undef

    const tuningPointA = tuner[a];
    const tuningPointB = tuner[b];
    const hueA = a > b ? wrapAround(tuningPointA.hue, [-180, 180]) : tuningPointA.hue;
    const hueB = tuningPointB.hue;

    if ((hueDoubleWrapped || hueWrapped) >= hueA && (hueDoubleWrapped || hueWrapped) < hueB) {
      // console.log('||| hue', hue, 'is between', tuningPointA.hue, 'and', tuningPointB.hue)
      let $ = hueDoubleWrapped || hueWrapped; // 70

      $ -= hueA; // 70 - 50 = 20 //

      $ /= hueB - hueA; // 20 / (120 - 50) = 2/7

      $ *= tuningPointB.sat - tuningPointA.sat; // -128 * 2 / 7 = -256 / 7 ~= -37

      $ += tuningPointA.sat;

      maxSat = $;
    }
  } // console.log('--- maxSat', maxSat)


  return maxSat;
});

var hueFromChannels = (({
  R,
  G,
  B
}) => {
  let hue;
  if (R > G && G >= B) hue = 60 * (0 + (G - B) / (R - B));
  if (G >= R && R > B) hue = 60 * (2 - (R - B) / (G - B));
  if (G > B && B >= R) hue = 60 * (2 + (B - R) / (G - R));
  if (B >= G && G > R) hue = 60 * (4 - (G - R) / (B - R));
  if (B > R && R >= G) hue = 60 * (4 + (R - G) / (B - G));
  if (R >= B && B > G) hue = 60 * (6 - (B - G) / (R - G));
  if (R === G && G === B) hue = 0; // console.log('||| found hue', hue)

  return hue;
});

var satFromChannels = (({
  R,
  G,
  B
}) => {
  const sat = Math.max(R, G, B) - Math.min(R, G, B); // console.log('||| found sat', sat)

  return sat;
});

var lumFromChannels = (({
  R,
  G,
  B
}) => {
  const lum = CHANNEL_SPECIFIC_LUM.R * R / 255 + CHANNEL_SPECIFIC_LUM.G * G / 255 + CHANNEL_SPECIFIC_LUM.B * B / 255; // console.log('||| found lum', lum)

  return lum;
});

var specToChannels = (({
  hue,
  sat,
  lum,
  prefer = 'sat',
  tuner
}) => {
  /*
  console.log('||| hue', hue)
  console.log('||| sat', sat)
  console.log('||| lum', lum)
  console.log('||| prefer', prefer)
  */
  const relativeChannels = hueToRelativeChannels(hue);

  function makeMinChannels(trueSaturation) {
    const makeMinChannel = idx => Math.round(relativeChannels[idx] * trueSaturation);

    return {
      R: makeMinChannel(0),
      G: makeMinChannel(1),
      B: makeMinChannel(2)
    };
  }

  let trueSaturation;
  let trueLuminosity;
  let minChannels;
  let minLum = 0;
  let maxLum = 1;
  let maxSat = maxSatForHueFromTuner(hue, tuner);

  if (prefer === 'sat') {
    trueSaturation = funnel(Math.min(sat, maxSat), [0, 255]);
    minChannels = makeMinChannels(trueSaturation);
    const maxChannels = {
      R: minChannels.R + 255 - trueSaturation,
      G: minChannels.G + 255 - trueSaturation,
      B: minChannels.B + 255 - trueSaturation
    };
    minLum = lumFromChannels(minChannels);
    maxLum = lumFromChannels(maxChannels);
    trueLuminosity = funnel(lum, [minLum, maxLum]);
    /*
    console.log('||| trueSaturation', trueSaturation)
    console.log('||| idealLum', lum)
    console.log('||| minChannels', minChannels)
    console.log('||| minLum', minLum)
    console.log('||| maxChannels', maxChannels)
    console.log('||| maxLum', maxLum)
    console.log('||| trueLuminosity', trueLuminosity)
    */
  }

  if (prefer === 'lum') {
    trueLuminosity = funnel(lum, [0, 1]);
    const specificLum = specificLumFromHue(hue);
    maxSat = Math.min(maxSat, Math.round(trueLuminosity <= specificLum ? 255 * (trueLuminosity / specificLum) : 255 * (1 - trueLuminosity) / (1 - specificLum)));
    trueSaturation = Math.min(sat, maxSat);
    minChannels = makeMinChannels(trueSaturation);
    minLum = lumFromChannels(minChannels);
    /*
    console.log('||| trueLuminosity', trueLuminosity)
    console.log('||| lumAtSat255', lumAtSat255)
    console.log('||| maxSat', maxSat)
    console.log('||| trueSaturation', trueSaturation)
    console.log('||| minChannels', minChannels)
    */
  }

  const maxWhite = 255 - Math.max(...Object.values(minChannels));
  const white = funnel(Math.round((trueLuminosity - minLum) * 255), [0, maxWhite]);
  const {
    R,
    G,
    B
  } = {
    R: minChannels.R + white,
    G: minChannels.G + white,
    B: minChannels.B + white
  };
  /*
  console.log('||| maxWhite', maxWhite)
  console.log('||| white', white)
  console.log('||| R', R)
  console.log('||| G', G)
  console.log('||| B', B)
  */

  return {
    channels: {
      R,
      G,
      B
    },
    fix: {
      sat: trueSaturation,
      lum: trueLuminosity
    },
    limit: {
      sat: [0, maxSat],
      lum: [prefer === 'lum' ? 0 : minLum, maxLum]
    }
  };
});

var channelsToHex = (channels => `#${Object.values(channels).map(channel => {
  let channelHex = channel.toString(16);
  if (channelHex.length === 1) channelHex = 0 + channelHex;
  return channelHex;
}).join('')}`);

var specToHexFixLimit = (({
  hue,
  sat,
  lum,
  prefer,
  tuner
}) => {
  const {
    channels,
    fix,
    limit
  } = specToChannels({
    hue,
    sat,
    lum,
    prefer,
    tuner
  });
  const {
    R,
    G,
    B
  } = channels;
  const hex = channelsToHex({
    R,
    G,
    B
  }); // console.log('--- newHex', hex)

  return {
    hex,
    fix,
    limit
  };
});

var specToHex = (({
  hue,
  sat,
  lum,
  prefer,
  tuner
}) => {
  const {
    hex
  } = specToHexFixLimit({
    hue,
    sat,
    lum,
    prefer,
    tuner
  });
  return hex;
});

function interpolateChanges(rootSpec, currentStep, axes) {
  const changes = {};

  for (let axisIdx = 0; axisIdx < axes.length; axisIdx++) {
    const currentAxis = axes[axisIdx];
    const {
      attribute,
      relative
    } = currentAxis;
    const rootValue = rootSpec[attribute];

    const to = () => {
      switch (typeof currentAxis.to) {
        case 'number':
          if (!relative) return currentAxis.to;
          return rootValue + currentAxis.to;

        case 'function':
          return currentAxis.to(rootSpec[attribute]);

        default:
          throw new Error(`gradient axis type error:\n` + `currentAxis.to is ${currentAxis.to} (type: ${typeof currentAxis.to})\n` + `type must be\n` + `* 'number'    (for absolute values) or\n` + `* 'function'  (for values relative to the root value).`);
      }
    };

    const from = () => {
      switch (typeof currentAxis.from) {
        case 'undefined':
          return rootSpec[attribute];

        case 'number':
          if (!relative) return currentAxis.from;
          return rootValue + currentAxis.from;

        case 'function':
          return currentAxis.from(rootSpec[attribute]);

        default:
          throw new Error(`gradient axis type error:\n` + `currentAxis.from is ${currentAxis.from} (type: ${typeof currentAxis.from})\n` + `type must be\n` + `* 'undefined' (to default to the root value)\n` + `* 'number'    (for absolute values) or\n` + `* 'function'  (for values relative to the root value).`);
      }
    };

    const value = interpolate({
      completionRatio: currentStep,
      range: [to(), from()]
    });
    changes[attribute] = value;
  } // console.log('>>> changes', changes)


  return changes;
}

function gradientToHexArray(rootSpec, gradient, tuner) {
  const {
    steps,
    axes,
    prefer
  } = gradient;
  const hexArray = [];

  for (let stepIdx = 0; stepIdx < steps.length; stepIdx++) {
    const currentStep = steps[stepIdx];
    const changes = interpolateChanges(rootSpec, currentStep, axes);
    const hex = specToHex(_extends({}, rootSpec, changes, {
      prefer,
      tuner
    }));
    hexArray.push(hex);
  }

  return hexArray;
}

var gradientsToHexArrays = ((originalColorObject, tuner) => {
  const {
    gradients
  } = originalColorObject;
  const {
    hue,
    sat,
    lum
  } = originalColorObject;
  const rootSpec = {
    hue,
    sat,
    lum
  };
  const hexArrays = [];

  for (let gradientIdx = 0; gradientIdx < gradients.length; gradientIdx++) {
    const gradient = gradients[gradientIdx];
    const hexArray = gradientToHexArray(rootSpec, gradient, tuner);
    hexArrays.push(hexArray);
  }

  return hexArrays;
});

var channelsToSpec = (({
  R,
  G,
  B
}) => {
  const hue = hueFromChannels({
    R,
    G,
    B
  });
  const sat = satFromChannels({
    R,
    G,
    B
  });
  const lum = lumFromChannels({
    R,
    G,
    B
  });
  return {
    hue,
    sat,
    lum
  };
});

var hexToChannels = (hex => {
  const validated = validateHex.process(hex);

  const channel = idx => parseInt(validated.substr(idx * 2 + 1, 2), 16);

  return {
    R: channel(0),
    G: channel(1),
    B: channel(2)
  };
});

var hexToSpec = (hex => {
  const {
    R,
    G,
    B
  } = hexToChannels(hex);
  /*
  console.log('+++ input hex', hex)
  console.log('||| R', R)
  console.log('||| G', G)
  console.log('||| B', B)
  */

  const {
    hue,
    sat,
    lum
  } = channelsToSpec({
    R,
    G,
    B
  });
  return {
    hue,
    sat,
    lum
  };
});

function mixPalettes({
  palettes,
  hexes,
  ops,
  place: [paletteIdx, stateKey, elementKey] = [0, null, null]
}) {
  if (!Array.isArray(palettes) || palettes.length === 0) throw new Error('Must provide at least one palette');
  let palettes_ = [...palettes];
  const palette = palettes_[paletteIdx];
  const state = stateKey ? palette.states[stateKey] : null;
  const element = elementKey ? state.elements[elementKey] : null;
  const destination = element || state || palette;
  const origin = element ? state : palette;
  if (!origin && hexes.length === 0) throw new Error('Must provide at least one hex');
  if (!origin.color) origin.color = hexToSpec(hexes[0]);

  let color = _extends({}, origin.color);

  const opsEntries = Object.entries(ops);
  /*
  console.log("--- from", { ...origin }.id, 'to', { ...destination }.id,
    `[${paletteIdx}, ${stateKey}, ${elementKey}]`,
  )
  console.log("    ||| origin", origin)
  console.log("    ||| hexes", hexes)
  console.log("    ||| color", color)
  console.log("    ||| ops", ops)
  console.log("    ||| palette", [...palette])
  */

  for (let idx = 0; idx < opsEntries.length; idx++) {
    const [key, value] = opsEntries[idx];
    /*
    console.log("    > color", ...color)
    console.log('     ', key)
    console.log('     ', value)
    */

    switch (key) {
      case 'color':
        // change colors
        // capture currentColor in palette
        switch (typeof value) {
          case 'number':
            color = hexToSpec(hexes[value]);
            break;

          case 'string':
            hexes.unshift(value);
            color = hexToSpec(value);
            break;

          default:
            throw new Error(`'color' accepts types 'string' or 'number'; got type '${typeof value}'`);
        }

        break;

      case 'hue':
      case 'sat':
      case 'lum':
        switch (typeof value) {
          case 'number':
            color[key] = value;
            break;

          case 'function':
            color[key] = value(color[key]);
            break;

          default:
            throw new Error(`hue, sat, lum accept types 'number' or 'function'; got type '${typeof value}'`);
        }

        break;

      case 'contrast':
        switch (value) {
          case 'hard':
            color.lum = color.lum > 0.67 ? 0 : 1;
            break;

          case 'soft':
            color.lum = color.lum > 0.67 ? 0.05 : 0.95;
            break;

          case 'harden':
            color.lum = color.lum > 0.67 ? 0 : 1;
            break;

          default:
            throw new Error(`'contrast' accepts 'hard' or 'soft'; got '${value}'`);
        }

        color.prefer = 'lum';
        break;

      case 'split':
      case 'trine':
      case 'tetra':
        break;

      case 'warm':
      case 'cool':
        break;

      case 'amp':
      case 'mute':
        break;

      case 'tint':
      case 'shade':
        const percentSaturated = color.sat / 255;
        const specificLum = specificLumFromHue(color.hue);
        const lightAdjust = 0.5 * (1 - percentSaturated) + specificLum * percentSaturated;
        const charge = key === 'tint' ? 100 : -100;
        const offset = lightAdjust * value / charge; // console.log('lightAdjust', lightAdjust)
        // console.log('offset', offset)

        color.lum += offset;
        if (color.lum < 0 || color.lum > 1) color.lum -= 2.5 * offset;
        color.prefer = 'lum';
        break;

      case 'holds':
        const holdsEntries = Object.entries(value);

        for (let holdsEntryIdx = 0; holdsEntryIdx < holdsEntries.length; holdsEntryIdx++) {
          const [subPaletteId, subPaletteOps] = holdsEntries[holdsEntryIdx];
          const externalElements = {};
          const baseElementsEntries = Object.entries(palette.states.base.elements);

          for (let baseElementsIdx = 0; baseElementsIdx < baseElementsEntries.length; baseElementsIdx++) {
            const [baseElementKey, baseElementValue] = baseElementsEntries[baseElementsIdx];
            externalElements[`ex-${baseElementKey}`] = baseElementValue;
          }

          palettes_.push({
            id: subPaletteId,
            color: palettes_[0].color,
            states: {
              base: {
                id: 'base',
                elements: externalElements
              }
            }
          });
          palettes_ = mixPalettes({
            palettes: palettes_,
            hexes,
            ops: subPaletteOps,
            place: [palettes_.length - 1, null, null]
          });
        }

        break;

      case 'state':
        const stateEntries = Object.entries(value);

        if (!element) {
          for (let stateEntryIdx = 0; stateEntryIdx < stateEntries.length; stateEntryIdx++) {
            const [stateKey_, stateOps] = stateEntries[stateEntryIdx]; // console.log('    ||| stateEntry', stateKey_, stateOps)

            palette.states[stateKey_] = {
              id: stateKey_,
              elements: {}
            };
            palette.states[stateKey_].color = mixPalettes({
              palettes,
              hexes,
              ops: stateOps,
              place: [paletteIdx, stateKey_, null]
            }); // console.log('      ', palette.states[stateKey_].hex)
          }
        } else {
          for (let stateEntryIdx = 0; stateEntryIdx < stateEntries.length; stateEntryIdx++) {
            const [stateKey_, stateOps] = stateEntries[stateEntryIdx]; // console.log('    ||| >>> stateEntry', stateKey_, stateOps)
            // console.log('    ||| >>> elementKey', elementKey)

            const state_ = palette.states[stateKey_];
            if (!state_.elements[elementKey]) state_.elements[elementKey] = {};
            palette.states[stateKey_].elements[elementKey].hex = mixPalettes({
              palettes,
              hexes,
              ops: stateOps,
              place: [paletteIdx, stateKey_, elementKey]
            });
          }
        }

        break;

      default:
        const elementKey_ = key;
        const stateKeys = Object.keys(palette.states);

        for (let stateKeysIdx = 0; stateKeysIdx < stateKeys.length; stateKeysIdx++) {
          const stateKey_ = stateKeys[stateKeysIdx];
          const state_ = palette.states[stateKey_]; // console.log('    ||| state_', state_)

          state_.elements[elementKey_] = {
            id: elementKey_
          };
          state_.elements[elementKey_].color = mixPalettes({
            palettes,
            hexes,
            ops: value,
            place: [paletteIdx, stateKey_, elementKey_]
          });
        }

    } // hex = specToHex({ ...color, tuner })


    destination.color = color;
    if (!state) destination.states.base.color = color; // console.log("    < color", color, color)
    // console.log("      set", { ...destination }.id)
  }

  if (element || state) {
    // console.log('---', destination.id, 'color', color)
    return color;
  } // console.log('--- returning palettes_', palettes_)


  return palettes_;
}

function paletteToScssBlock(palette, tuner) {
  let scssBlock = '';
  const baseEntries = Object.entries(palette.states.base.elements);

  for (let baseIdx = 0; baseIdx < baseEntries.length; baseIdx++) {
    const [key, element] = baseEntries[baseIdx];
    scssBlock += `  --${key}-color: ${specToHex(_extends({}, element.color, {
      tuner
    }))};\n`;
  }

  const stateEntries = Object.entries(palette.states);

  for (let stateIdx = 0; stateIdx < stateEntries.length; stateIdx++) {
    const [stateKey, stateObj] = stateEntries[stateIdx]; // console.log('stateKey', stateKey)

    switch (stateKey) {
      case 'hover':
        scssBlock += `  &:hover, &:focus, &:focus-within {\n`;
        break;

      case 'active':
        scssBlock += `  &:active, &.active {\n`;
        break;

      case 'disabled':
        scssBlock += `  &:disabled, &.disabled {\n`;
        break;

      default:
        continue;
    }

    const elementEntries = Object.entries(stateObj.elements);

    for (let elementIdx = 0; elementIdx < elementEntries.length; elementIdx++) {
      const [key, element] = elementEntries[elementIdx];
      scssBlock += `    --${key}-color: ${specToHex(_extends({}, element.color, {
        tuner
      }))};\n`;
    }

    scssBlock += '  }\n';
  }

  return scssBlock;
}

function makeScssObject(palettes, tuner) {
  const scssObject = {};

  for (let index = 0; index < palettes.length; index++) {
    const palette = palettes[index];
    scssObject[palette.id] = paletteToScssBlock(palette, tuner);
  }

  return scssObject;
}

const paletteTemplate = name => ({
  id: name,
  states: {
    base: {
      elements: {}
    }
  }
});

function makeColorPalettesFromHexes({
  hexes,
  scheme,
  tuner
}) {
  const schemeEntries = Object.entries(scheme);
  let palettes = [];

  for (let idx = 0; idx < schemeEntries.length; idx++) {
    const [name, ops] = schemeEntries[idx];
    const newPalettes = mixPalettes({
      palettes: [paletteTemplate(name)],
      hexes,
      ops
    });
    palettes = [...palettes, ...newPalettes];
  }

  const scss = makeScssObject(palettes, tuner); // console.log(scss)

  return scss;
}

var getStandout = (hex => {
  const {
    lum
  } = hexToSpec(hex);
  const standOut = lum > 0.666 ? 'black' : 'white';
  return standOut;
}); // input: red
// hard standout: white

var getOffset = (({
  hex,
  tuner,
  offsets
}) => {
  const rootSpec = hexToSpec(hex);
  const changes = {};

  for (let index = 0; index < offsets.length; index++) {
    const {
      attribute,
      offsetValue
    } = offsets[index];
    const rootValue = rootSpec[attribute];
    const {
      sat
    } = rootSpec;
    let newValue;

    switch (attribute) {
      case 'hue':
      case 'sat':
        newValue = rootValue + offsetValue;
        break;

      case 'lum':
        if (offsetValue + rootValue < 0 || offsetValue + rootValue > 1) {
          newValue = rootValue - (offsetValue - (sat >= 170 ? offsetValue * (sat - 170) / 85 : 0)) * 1.67;
        } else {
          newValue = rootValue + (offsetValue - (sat >= 170 ? 0.333 * offsetValue * (sat - 170) / 85 : 0));
        }

        break;

      default:
        throw new Error('Received spec attribute other than "hue," "sat," or "lum."');
    }

    changes[attribute] = newValue;
  }

  const newHex = specToHex(_extends({}, rootSpec, changes, {
    prefer: 'lum',
    tuner
  }));
  return newHex;
});

function hardContrast(hex) {
  const {
    lum
  } = hexToSpec(hex);
  const contrast = lum > 0.666 ? 'black' : 'white';
  return contrast;
}
function softContrast(hex, tuner) {
  const rootSpec = hexToSpec(hex);
  const newLum = rootSpec.lum > 0.666 ? 0.05 : 0.95;
  const newHex = specToHex(_extends({}, rootSpec, {
    lum: newLum,
    prefer: 'lum',
    tuner
  }));
  return newHex;
} // input: red
// hard standout: white

const rangeNames = [{
  hue: 20,
  name: 'orange'
}, {
  hue: 45,
  name: 'yellow'
}, {
  hue: 61,
  name: 'citron'
}, {
  hue: 80,
  name: 'lime'
}, {
  hue: 100,
  name: 'green'
}, {
  hue: 150,
  name: 'teal'
}, {
  hue: 180,
  name: 'cyan'
}, {
  hue: 200,
  name: 'blue'
}, {
  hue: 240,
  name: 'indigo'
}, {
  hue: 270,
  name: 'violet'
}, {
  hue: 290,
  name: 'magenta'
}, {
  hue: 330,
  name: 'pink'
}, {
  hue: 350,
  name: 'red'
}];
var nameHue = (hue => {
  console.log('||| hue', hue);
  const hueWrapped = wrapAround(hue, [0, 360]);

  for (let a = -1, b = 0; b < rangeNames.length; a++, b++) {
    a = wrapAround(a, [0, rangeNames.length]);
    const hueDoubleWrapped = a > b ? wrapAround(hueWrapped, [-180, 180]) : undefined;
    const namingPointA = rangeNames[a];
    const namingPointB = rangeNames[b];
    const hueA = a > b ? wrapAround(namingPointA.hue, [-180, 180]) : namingPointA.hue;
    const hueB = namingPointB.hue;

    if ((hueDoubleWrapped || hueWrapped) >= hueA && (hueDoubleWrapped || hueWrapped) < hueB) {
      console.log('||| hue', hue, 'is between', namingPointA.hue, 'and', namingPointB.hue, 'therefore it is named', namingPointA.name);
      return namingPointA.name;
    }
  }
});

var builtInTunerKit = {
  noFilter: [{
    sat: 255,
    hue: 0
  }, {
    sat: 255,
    hue: 360
  }],
  simulateCMYK: [{
    sat: 255,
    hue: 55
  }, {
    sat: 127,
    hue: 120
  }, {
    sat: 255,
    hue: 200
  }, {
    sat: 255,
    hue: 205
  }, {
    sat: 170,
    hue: 240
  }, {
    sat: 230,
    hue: 270
  }, {
    sat: 210,
    hue: 300
  }, {
    sat: 255,
    hue: 320
  }]
};

var schemes = {
  trifactory: {
    PaletteModule: {
      color: 0,
      bg: {},
      mg: {
        shade: 10
      },
      fg: {
        contrast: 'soft'
      },
      holds: {
        Control: {
          state: {
            hover: {},
            active: {},
            disabled: {}
          },
          bg: {
            shade: 24,
            state: {
              base: {},
              hover: {
                shade: 18
              },
              active: {
                shade: 32
              },
              disabled: {
                shade: 32
              }
            }
          },
          mg: {
            shade: 48
          },
          fg: {
            contrast: 'soft',
            state: {
              hover: {
                contrast: 'harden'
              },
              active: {
                contrast: 'harden'
              },
              disabled: {
                shade: 20
              }
            }
          }
        }
      }
    }
  }
}; // Example:
/*
const colorSchema = {
  exfg: colorContext.softContrast,
  exbg: hex,
  fg: {
    base: colorContext.softContrast,
    hover: colorContext.softContrast,
    active: colorContext.hardContrast,
    disable: a,
  },
  mg: [20, 15, 25].map(n => shade(n)),
  bg: [10, 5, 15].map(n => shade(n)),
}

const stylesheet = `
a
`
*/

exports.CHANNEL_SPECIFIC_LUM = CHANNEL_SPECIFIC_LUM;
exports.HUE_STRUCTURES = HUE_STRUCTURES;
exports.builtInTunerKit = builtInTunerKit;
exports.channelsToSpec = channelsToSpec;
exports.funnel = funnel;
exports.getOffset = getOffset;
exports.getStandout = getStandout;
exports.gradientsToHexArrays = gradientsToHexArrays;
exports.hardContrast = hardContrast;
exports.hexToChannels = hexToChannels;
exports.hexToSpec = hexToSpec;
exports.hueFromChannels = hueFromChannels;
exports.hueToRelativeChannels = hueToRelativeChannels;
exports.interpolate = interpolate;
exports.lumFromChannels = lumFromChannels;
exports.maxSatForHueFromTuner = maxSatForHueFromTuner;
exports.nameHue = nameHue;
exports.satFromChannels = satFromChannels;
exports.scheme = makeColorPalettesFromHexes;
exports.schemes = schemes;
exports.softContrast = softContrast;
exports.specToHex = specToHex;
exports.specToHexFixLimit = specToHexFixLimit;
exports.specificLumFromHue = specificLumFromHue;
exports.validateHex = validateHex;
exports.wrapAround = wrapAround;
//# sourceMappingURL=luum.js.map
